﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text.RegularExpressions;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Xml;
using System.Xml.Linq;
using Windows.Foundation;

namespace Office365AuthHelperRT
{

    public sealed class Office365AuthHelperRT
    {

        #region Constants
        private const string office365STS = "https://login.microsoftonline.com/extSTS.srf";
        //private const string office365Login = "https://login.microsoftonline.com/login.srf";
        private const string wsse = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";
        private const string wsu = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd";
        //private const string userAgent = "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)";
        #endregion

        private Cookie rtFaCookie;
        private Cookie fedAuthCookie;
        private CookieContainer cookieContainer = new CookieContainer();



        public IAsyncOperation<String> executeQuery(String url, String soapMessage, String soapAction)
        {
            return Task.Run<String>(async () =>
            {
                HttpClientHandler handler = new HttpClientHandler();
                handler.UseDefaultCredentials = true;
                handler.AllowAutoRedirect = false;
                //important! add the required cookies (rtDa, FedAuth) to each request
                handler.CookieContainer = this.cookieContainer;

                HttpClient client = new HttpClient(handler);

                HttpContent httpContent = new StringContent(soapMessage);
                httpContent.Headers.Add("SOAPAction", soapAction);
                httpContent.Headers.ContentLength = soapMessage.Length;
                httpContent.Headers.ContentType = MediaTypeHeaderValue.Parse("text/xml;charset=utf-8");// ("application/x-www-form-urlencoded");

                HttpResponseMessage response = await client.PostAsync(url, httpContent);
                string returnContent = await response.Content.ReadAsStringAsync();

                return returnContent;
            }).AsAsyncOperation();
        }

        public IAsyncAction signin(string username, string password, string host)
        {
            return Task.Run(async () =>
            {

                //request the security token from the server
                var tokenReply = await geTokenFromSPOAsync(getSecurityTokenXml(username, password, host));

                // parse the token response
                XDocument doc = XDocument.Parse(tokenReply);
                // get the security token
                var crypt = from result in doc.Descendants()
                            where result.Name == XName.Get("BinarySecurityToken", wsse)
                            select result;
                // get the token expiration
                var expires = from result in doc.Descendants()
                              where result.Name == XName.Get("Expires", wsu)
                              select result;

                //get cookies based on token
                string token = crypt.FirstOrDefault().Value;
                var headers = await getHeadersFromSPOAsync(host+"/_forms/default.aspx?wa=wsignin1.0", token);
                foreach (KeyValuePair<string, string> pair in headers)
                {
                    if (pair.Key.Equals("Set-Cookie"))
                    {
                        this.rtFaCookie = new Cookie("rtFa", extractCookieVal(extractCookie(pair.Value, "rtFa")));
                        this.rtFaCookie.Expires = Convert.ToDateTime(expires.First().Value);
                        this.fedAuthCookie = new Cookie("FedAuth", extractCookieVal(extractCookie(pair.Value, "FedAuth")));
                        this.fedAuthCookie.Expires = Convert.ToDateTime(expires.First().Value);
                        this.cookieContainer.Add(new Uri(host), this.rtFaCookie);
                        this.cookieContainer.Add(new Uri(host), this.fedAuthCookie);
                    }
                }

            }).AsAsyncAction();
        }




        private async Task<IDictionary<string, string>> getHeadersFromSPOAsync(string targetUri, string content)
        {
            HttpClientHandler handler = new HttpClientHandler();
            handler.UseDefaultCredentials = true;
            handler.AllowAutoRedirect = false;

            HttpClient client = new HttpClient(handler);

            HttpContent httpContent = new StringContent(content);
            httpContent.Headers.ContentType = new MediaTypeHeaderValue("application/x-www-form-urlencoded");

            HttpResponseMessage response = await client.PostAsync(targetUri, httpContent);

            Dictionary<string, string> headers = new Dictionary<string, string>();

            foreach (var header in response.Headers)
            {
                string headerValue = string.Empty;

                if (header.Value.Count() == 1)
                {
                    headerValue = header.Value.First();
                }
                else
                {
                    StringBuilder stringValue = new StringBuilder();

                    foreach (var value in header.Value)
                    {
                        stringValue.Append(value);
                        stringValue.Append("\n");
                    }

                    headerValue = stringValue.ToString();
                }

                headers.Add(header.Key, headerValue);
            }

            return headers;
        }



        private async Task<String> geTokenFromSPOAsync(string content)
        {
            HttpClientHandler handler = new HttpClientHandler();
            handler.UseDefaultCredentials = true;
            handler.AllowAutoRedirect = false;

            HttpClient client = new HttpClient(handler);

            HttpContent httpContent = new StringContent(content);
            httpContent.Headers.ContentType = new MediaTypeHeaderValue("application/x-www-form-urlencoded");

            HttpResponseMessage response = await client.PostAsync(office365STS, httpContent);
            string returnContent = await response.Content.ReadAsStringAsync();

            return returnContent;
        }


        /*
         * Helper method used to extract a specific cookie out of a set-cookie string
         */
        private String extractCookie(String cookieString, String cookieName)
        {
            Regex r = new Regex(String.Format(@"({0})=([^\s]+)", cookieName));
            String cookie = "";
            foreach (Match m in r.Matches(cookieString))
            {
                cookie += m;
            }
            char[] trimChar = { ';' };
            return cookie.Trim(trimChar);
        }

        /*
         * helper method used to extract the value of a cookie
         */
        private String extractCookieVal(String cookie)
        {
            return cookie.Substring(cookie.IndexOf("=") + 1);
        }


        /**
         * creates the SOAP envelope needed to fetch the Security Token
         * 
         */
        private String getSecurityTokenXml(String username, String password, String host)
        {
            return "<s:Envelope xmlns:s=\"http://www.w3.org/2003/05/soap-envelope\" xmlns:a=\"http://www.w3.org/2005/08/addressing\" xmlns:u=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd\">" +
              "<s:Header>" +
                "<a:Action s:mustUnderstand=\"1\">http://schemas.xmlsoap.org/ws/2005/02/trust/RST/Issue</a:Action>" +
                "<a:ReplyTo>" +
                  "<a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>" +
                "</a:ReplyTo>" +
                "<a:To s:mustUnderstand=\"1\">https://login.microsoftonline.com/extSTS.srf</a:To>" +
                "<o:Security s:mustUnderstand=\"1\" xmlns:o=\"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd\">" +
                  "<o:UsernameToken>" +
                    "<o:Username>" + username + "</o:Username>" +
                    "<o:Password>" + password + "</o:Password>" +
                  "</o:UsernameToken>" +
                "</o:Security>" +
              "</s:Header>" +
              "<s:Body>" +
                "<t:RequestSecurityToken xmlns:t=\"http://schemas.xmlsoap.org/ws/2005/02/trust\">" +
                  "<wsp:AppliesTo xmlns:wsp=\"http://schemas.xmlsoap.org/ws/2004/09/policy\">" +
                    "<a:EndpointReference>" +
                      "<a:Address>" + host + "</a:Address>" +
                    "</a:EndpointReference>" +
                  "</wsp:AppliesTo>" +
                  "<t:KeyType>http://schemas.xmlsoap.org/ws/2005/05/identity/NoProofKey</t:KeyType>" +
                  "<t:RequestType>http://schemas.xmlsoap.org/ws/2005/02/trust/Issue</t:RequestType>" +
                  "<t:TokenType>urn:oasis:names:tc:SAML:1.0:assertion</t:TokenType>" +
                "</t:RequestSecurityToken>" +
              "</s:Body>" +
            "</s:Envelope>";
        }
    }
}
